/*Kamil bikowski - gr H5ISI - PRACA DOMOWA Z PRZEDMIOTU ZPR			|
|_______________________________________________________________________________|
|		Fabryka Obiektw - rejestracja metod fabrycznych		|
|										|
|Problem: 	Tworzenie obiektu jest w jezyku c++ statytyczne. Nie zawiera on |
|		mechanizmu tworzenia obiektw w sposb dynamiczny. Pojcia klasy|
|		i obiektu s rne. Klasa jest tworzona przez programist - jest|
|		definicj obiektu, ktry to z kolei jest tworzony w trakcie	|
|		dziaania programu. Czsto jednak pojawia si potrzeba aby 	|
|		w sposb dynamiczny generowa obiekty (np. odczyt/zapis obiektw|
|		na dysk) albo np. pozostawi klientom moliwo modyfikacji	|
|		zachowania obiektw za pomoc polimorfizmu poprzez wywoanie  	|
|		funkcji wirtualnej Create zdefiniowanej przez klas wysokiego 	|
|		poziomu. Rozwizanie "siowe" poprzez zastosowanie instrukcji	|
|		switch nie jest najlepszym rozwizaniem:			|
|			- mija sie z ide obiektowoci				|
|			- w jednym pliku rdowym jest zawarta informacja	|
|			  o wszystkich podtypach co czyni go wskim gardem,	|
|			  jeli chodzi o zalenoci kompilacji i utrzmanie kodu.|
|			- modyfiukujc kod musimy go zmienia w wielu miejscach |
|Rozwiznie:	Naprzeciw powyszemu problemowi wychodzi wzorzec fabryki 	|
|		obiektw z rejestracj metod fabrycznych, ktry w przyblieniu	|
|		wyglda obrazuje poniszy schemat:				|
|			 _______		 _______________________	|	
|			|Produkt|<--------------|     FabrykaObiektow	|	|
|			|_______|		|-----------------------|	|
|			    ^			|+ StworzProdukt()	|	|
|			    |	 		|+ RejestrujProduct()	|	|
|			    |			|_______________________|	|
|			 ___|_____						|
|			|Produkt  |						|	
|			|Konkretny|						|
|			|_________|						|
|										|
|		Produkt konkretny - fabryka dostarcza produkt w postaci obiektu |
|		Produkt - abstrakcyjny typ podstawowy. Fabryka zwraca wskanik	|
|			  do tego obiektu nie przekazujc informacji o produkcie|
|			  konkretnym.						|
|		Informacje o konkretnym produkcie przechowywane s przez fabryk|
|		w mapie (np. std::map). Kady konkretny produkt ma przydzielony	|
|		unikalny identyfiaktor na podstawie, ktrego jest identyfikowany|
|		Wasciwie jest przechowywany wskanik do funkcji zwracajcej 	|
|		wskanik do Produkt.						|
|		Rejestracja kolejnego produktu konkretnego odbywa si poprzez	|
|		podanie unikalnego identyfikatora danego Produktu Konkretnego	|
|		oraz wskanika na funkcj zwracajc wskanik na Produkt. 	|
|		Dziki zastosowaniu polimorfizmu wyeliminowalimy wady 		|
|		rozwizania siowego. Teraz dodanie kolejnego Produktu 		|
|		Konkretnego sprowadza si do stworzenia nowej klasy definiujcej|
|		go, zawierajcej metod postaci:				|
|			Produkt* StworzKonkretnyProdukt(){ 			|
|				return new ProduktKonkretny;			|
|			}							|
|		Rejestracja natomiast polega na wywolaniu RegisterProduct na 	|
|		rzecz wyej wymienionej metody. Od tej chwili stworzenie	|
|		konkretnego produktu bdzie polegao na wywoaniu metody 	|
|		StworzProdukt z okrlonym id podanym w porcesie rejstracji	|
|		Czesto fabryka obiektw powinna by singletonem, ale to juz jest|
|		temat na oddzieln prac.
|______________________________________________________________________________*/
#include <iostream>
#include <map>

/**
  * Szablon fabryki. Product - klasa abstrakcyjan, Id - identyfiaktor klasy, ProductCreator - 
  * wskanik na funkcj zwracajc wskanik na obiekt konkretny dziedziczcy po Product
  */
template <class Product, class Id, class ProductCreator = Product* (*)()>
class Factory{
	public:
		//Metoda produkujca konkretne obiekty na podstawie ich identyfiaktorw.
		Product* CreateProduct(const Id& id){
			typename Map::const_iterator i = associations.find(id);
			if(i != associations.end()){
				return (i->second());  //jeli w mapie jest podany w parametrze klucz
						       //wywoaj powizan z nim funkcj
			}
			else
				return NULL; //obsuga bdw - przykadowe rozwizanie
					     //mona zaimplementowa rozwizanie z uyciem wyjtkw
		}
		//Metoda rejestrujca obiekty w fabryce.
		bool Register(const Id& id, ProductCreator creator){
			return associations.insert(
					typename Map::value_type(id, creator)).second;
		}
		//Metoda wyrejestrowujca obiekty.
		bool UnRegister(const Id& id){
			return associations.erase(id) == 1;
		}
	private:
		typedef std::map<Id, ProductCreator> Map; 
		Map associations; //mapa suy do poczenia identyfiatora z danym wsknikiem na funkcj
				  //wymagane jest aby Id miao zaimplementowane operatory '==', '<'
				  //
};
//Klasa abstrakcyjna Ford. Diedzicz po niej konkretne modele samochodw, ktre rejestruj si w fabryce
//a nastpnie s przez ni produkowane.
class Ford{
	public:
		//funkcja, ktra w kadej z klas dziedziczcych po Ford 
		//wywietla rodzaj danego obiektu
		virtual void Info() = 0;
};
Factory<Ford, int> FordFactory; //obiekt relizujcy fabryk samochodw marki Ford
//Kada z poniszych przestrzeni nazw zawierajca konkretne implementacje 
//klasy Ford moe znajdowa si w osobnym pliku. Chcc doda kolejn mark samochodu
//dodajemy kolejny plik z zawartoci podobn jak kady z namespace'ow. 
namespace Focus{
	static const int ID = 1;
	class FocusFord : public Ford{
		public:
			virtual void Info(){
				std::cout << "Jestem Fordem Focus"<<std::endl;
			}
		private:
			//konkretne waciwoci danego modelu
	};
	Ford* CreateFocusFord(){
		return new FocusFord;
	}
	void StartProduce(){
		FordFactory.Register(Focus::ID, &Focus::CreateFocusFord);
	}
	void StopProduce(){
		FordFactory.UnRegister(Focus::ID);
	}
}
namespace Galaxy{
	static const int ID = 2;
	class GalaxyFord : public Ford{
		public:
			virtual void Info(){
				std::cout << "Jestem Fordem Galaxy"<<std::endl;
			}
		private:
			//konkretne waciwoci danego modelu
	};
	Ford* CreateGalaxyFord(){
		return new GalaxyFord;
	}
	void StartProduce(){
		FordFactory.Register(Galaxy::ID, &Galaxy::CreateGalaxyFord);
	}
	void StopProduce(){
		FordFactory.UnRegister(Galaxy::ID);
	}
}
namespace Mondeo{
	static const int ID = 3;
	class MondeoFord : public Ford{
		public:
			virtual void Info(){
				std::cout << "Jestem Fordem Mondeo"<<std::endl;
			}
		private:
			//konkretne waciwoci danego modelu
	};
	Ford* CreateMondeoFord(){
		return new MondeoFord;
	}
	void StartProduce(){
		FordFactory.Register(Mondeo::ID, &Mondeo::CreateMondeoFord);
	}
	void StopProduce(){
		FordFactory.UnRegister(Mondeo::ID);
	}
}
//Przyklady uycia fabryki
int main(){
	//Trzy wskaniki na obiekty typu Ford. Bed si w nich znajdoway wskazania na konkretne klasy pochodne klasy Ford.
	Ford* any1; 
	Ford* any2; 
	any1 = FordFactory.CreateProduct(Focus::ID); //Prbujemy wyprodukowa Forda Focus'a
	if(any1)
		any1->Info();	//jeli si udao to wywietl informacje o samochodzie
				//oczywicie w tym wypadku operacja si nie uda, poniewa nie 
				//zarejestrowalimy tego podtypu w fabryce
	else
		//jeli nie to wypisz stosowny komunikat
		std::cout<<"Fabryka nie produkuje tego typu samochodu"<<std::endl;
	//rejestrujemy podtyp w fabryce
	Focus::StartProduce();
	any1 = FordFactory.CreateProduct(Focus::ID);
	if(any1)
		any1->Info();	//jeli si udao to wywietl informacje o samochodzie, teraz ju dokonalimy rejestracji
	else
		//jeli nie to wypisz stosowny komunikat
		std::cout<<"Fabryka nie produkuje tego typu samochodu"<<std::endl;
	//Zaczynamy produkowa modele Galaxy
	Galaxy::StartProduce();
	any2 = FordFactory.CreateProduct(Galaxy::ID); //produkujemy 1 egzemplarz Galaxy
	if(any2)
		any2->Info();	//jeli si udao to wywietl informacje o samochodzie, teraz ju dokonalimy rejestracji
	else
		//jeli nie to wypisz stosowny komunikat
		std::cout<<"Fabryka nie produkuje tego typu samochodu"<<std::endl;
	Galaxy::StopProduce();
	Focus::StopProduce();
	//staramy sie wyprodukowa Galaxy i Focus po wyrejestrowaniu ich z fabryki
	any1 = FordFactory.CreateProduct(Focus::ID);
	any2 = FordFactory.CreateProduct(Galaxy::ID);
	if(any2)
		any2->Info();	//jeli si udao to wywietl informacje o samochodzie, teraz ju dokonalimy rejestracji
	else
		//jeli nie to wypisz stosowny komunikat
		std::cout<<"Fabryka nie produkuje tego typu samochodu"<<std::endl;
	if(any1)
		any2->Info();	//jeli si udao to wywietl informacje o samochodzie, teraz ju dokonalimy rejestracji
	else
		//jeli nie to wypisz stosowny komunikat
		std::cout<<"Fabryka nie produkuje tego typu samochodu"<<std::endl;
	return 0;
}
